#!/usr/bin/python

######################################
#    Copyright 2009 Tamara Broderick
#    This file is part of Nomon Draw.
#
#    Nomon Draw is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Nomon Draw is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Nomon Draw.  If not, see <http://www.gnu.org/licenses/>.
######################################

import broderclocks
import config
#import glob
import kconfig
import string
import time
import Tkinter
import numpy
if(kconfig.target_evt==kconfig.joy_evt):
    import pygame
from ToolTip import *
import sys
import random
import os

class Keyboard:
    def __init__(self, parent):
        ### process the inputs ###
        ## copy
        self.parent = parent

        ## command line
        # check that number of arguments is valid
        if len(sys.argv) < 2:
            self.setup = 1
        else:
            # read arguments
            self.setup = string.atoi(sys.argv[1])

        ### initialize ###
        self.Nclocks = kconfig.sqrt_Nclocks*kconfig.sqrt_Nclocks

        ## initialize pygame and joystick
        if(kconfig.target_evt is kconfig.joy_evt):
            pygame.init()
            if(pygame.joystick.get_count() < 1):
                # no joysticks found
                print "Please connect a joystick.\n"
                self.quit(None)
            else:
                # create a new joystick object from
                # ---the first joystick in the list of joysticks
                Joy0 = pygame.joystick.Joystick(0)
                # tell pygame to record joystick events
                Joy0.init()
            ## start looking for events
            self.parent.after(0, self.find_events)
        ## not in selection pause
        self.in_pause = False

        ## determine clock positions
        self.init_locs() # free-draw
	self.init_controls() # controls
        ## set up canvas
        self.gen_canvas()
	self.gen_side_canvas()
	self.gen_controls()
	self.gen_speed_scale()
	self.gen_buttons()
	## set up drawing
	self.old_index = -1
	self.init_draw()
        ## set up "talked" text
        self.talk_file = "talk.txt"
        ## set up broderclocks
        self.file_handle = None
        self.bc = broderclocks.BroderClocks(self, self.canvas, self.clock_centers, self.win_diffs, kconfig.clock_rad, self.file_handle, self.words_on, self.words_off, config.bgcolor, time.time(), 0, 0, self.time_rotate, None)
        self.wait_s = self.bc.get_wait()
        self.bc.init_follow_up(self.word_score_prior)
	# histogram
	self.gen_histogram()

        ### animate ###
        self.last_anim_call = self.canvas.after(0, self.on_timer)

    def find_events(self):
        ## check everything in the queue of pygame events
        events = pygame.event.get()
        for event in events:
            # event type for pressing any of the joystick buttons down
            if event.type == pygame.JOYBUTTONDOWN:
                # generate the event I've defined
                self.canvas.event_generate(kconfig.joy_evt)

        ## return to check for more events in a moment
        self.parent.after(20, self.find_events)

    def init_draw(self):
	# last points in current object
	self.loc_inds = []
	# all things drawn so far
	self.drawn = []
	self.side_drawn = []

    def init_locs(self):
        w_block = kconfig.w_block
        h_block = kconfig.h_block
        self.w_canvas = w_block * (kconfig.sqrt_Nclocks + 0.5)
        self.h_canvas = h_block * (kconfig.sqrt_Nclocks + 0.5)
    
        self.clock_centers = []
        self.win_diffs = []
        self.rect_locs = []
        self.labels = []
        self.words_on = []
        self.words_off = []
        self.word_score_prior = []
        self.index = 0
        for nh in range(0,kconfig.sqrt_Nclocks):
            for nw in range(0,kconfig.sqrt_Nclocks):
                if self.setup is 2:
                    x = random.randint(0,self.w_canvas)
                    y = random.randint(0,self.h_canvas)
                else:
                    x = nw*w_block+kconfig.clock_rad + 0.5*w_block
                    y = nh*h_block+kconfig.clock_rad + 0.5*h_block
                self.clock_centers.append([x,y])
                self.labels.append("Selected pixel")
                # rectangles
                self.rect_locs.append([x-0.5*w_block,y-0.5*h_block,x+0.5*w_block,y+0.5*h_block])
                # winning differences
                self.win_diffs.append(config.win_diff_base)
                # any prior probability weights
                self.word_score_prior.append(0)
                # which clocks to display
                self.words_on.append(self.index)
                self.index = self.index + 1

    def init_controls(self):
	self.w_controls = self.w_canvas/2.0
	self.h_control = self.h_canvas/2/5
	self.x_controls = self.w_canvas
	self.y_controls = self.h_canvas/2.0

	self.control_text = []
	self.control_text_locs = []
	self.control_index = 0

	### pen up/down
	self.pen_down = True
	self.line_fill = kconfig.line_fill
	# clock
	x = self.x_controls
	y = self.y_controls
	self.clock_centers.append([x + 2*kconfig.clock_rad,y + self.h_control/2])
	# text
	self.control_text.append("Start Line Series")
	self.control_text_locs.append([x + 4*kconfig.clock_rad, y + self.h_control/2])
	# rectangle to highlight
	self.rect_locs.append([x, y, x + self.w_controls, y + self.h_control])
        # winning differences
        self.win_diffs.append(config.win_diff_base)
        # any prior probability weights
        self.word_score_prior.append(0)
        # which clocks to display
	self.pen_toggle_clock_ind = self.index
	self.pen_toggle_control_ind = self.control_index
        self.words_on.append(self.index)
	self.labels.append("Toggled line")
      	self.index = self.index + 1
	self.control_index = self.control_index + 1

	### rectangle
	self.in_rectangle = False
	# clock
	y = y + self.h_control
	self.clock_centers.append([x + 2*kconfig.clock_rad,y + self.h_control/2])
	# text
	self.control_text.append("Start Rectangle")
	self.control_text_locs.append([x + 4*kconfig.clock_rad, y + self.h_control/2])
	# rectangle to highlight
	self.rect_locs.append([x, y, x + self.w_controls, y + self.h_control])
        # winning differences
        self.win_diffs.append(config.win_diff_base)
        # any prior probability weights
        self.word_score_prior.append(0)
        # which clocks to display
	self.rectangle_clock_ind = self.index
	self.rectangle_control_ind = self.control_index
        self.words_on.append(self.index)
	self.labels.append("Drawing rectangle")
      	self.index = self.index + 1
	self.control_index = self.control_index + 1

	### oval
	self.in_oval = False
	# clock
	y = y + self.h_control
	self.clock_centers.append([x + 2*kconfig.clock_rad,y + self.h_control/2])
	# text
	self.control_text.append("Start Oval")
	self.control_text_locs.append([x + 4*kconfig.clock_rad, y + self.h_control/2])
	# rectangle to highlight
	self.rect_locs.append([x, y, x + self.w_controls, y + self.h_control])
        # winning differences
        self.win_diffs.append(config.win_diff_base)
        # any prior probability weights
        self.word_score_prior.append(0)
        # which clocks to display
	self.oval_clock_ind = self.index
	self.oval_control_ind = self.control_index
        self.words_on.append(self.index)
	self.labels.append("Drawing oval")
      	self.index = self.index + 1
	self.control_index = self.control_index + 1

	### fill on/off
	self.fill_on = False
	self.shape_fill = ""
	# clock
	y = y + self.h_control
	self.clock_centers.append([x + 2*kconfig.clock_rad,y + self.h_control/2])
	# text
	self.control_text.append("Turn On Fill")
	self.control_text_locs.append([x + 4*kconfig.clock_rad, y + self.h_control/2])
	# rectangle to highlight
	self.rect_locs.append([x, y, x + self.w_controls, y + self.h_control])
        # winning differences
        self.win_diffs.append(config.win_diff_base)
        # any prior probability weights
        self.word_score_prior.append(0)
        # which clocks to display
	self.fill_toggle_clock_ind = self.index
	self.fill_toggle_control_ind = self.control_index
        self.words_on.append(self.index)
	self.labels.append("Toggled fill")
      	self.index = self.index + 1
	self.control_index = self.control_index + 1

	### delete
	# clock
	y = y + self.h_control
	self.clock_centers.append([x + 2*kconfig.clock_rad,y + self.h_control/2])
	# text
	self.control_text.append("Undo Last")
	self.control_text_locs.append([x + 4*kconfig.clock_rad, y + self.h_control/2])
	# rectangle to highlight
	self.rect_locs.append([x, y, x + self.w_controls, y + self.h_control])
        # winning differences
        self.win_diffs.append(config.win_diff_base)
        # any prior probability weights
        self.word_score_prior.append(0)
        # which clocks to display
	self.delete_clock_ind = self.index
	self.delete_control_ind = self.control_index
        self.words_on.append(self.index)
	self.labels.append("Deleting last item drawn")
      	self.index = self.index + 1
	self.control_index = self.control_index + 1

    def gen_canvas(self):
        self.canvas = Tkinter.Canvas(self.parent, width=self.w_canvas + self.w_controls, height=self.h_canvas, background=config.bgcolor)
        self.canvas.grid(row=0,column=0,columnspan=2)
        self.canvas.bind(kconfig.target_evt, self.on_press)
        self.canvas.focus_set() # so don't have to click on the canvas beforehand

	# clock space
	self.draw_canvas = self.canvas.create_rectangle((0,0,self.w_canvas,self.h_canvas),fill="",outline="black")

	# keys for coloring on wins
	self.key_id = []
        for index in range(0,self.Nclocks):
            self.key_id.append(self.canvas.create_rectangle(self.rect_locs[index],fill=kconfig.key_color,outline=""))

	self.keyboard_tooltip = ToolTip(self.canvas, follow_mouse=1, font=kconfig.word_font, text="This is the Nomon Draw canvas. To select a pixel, find the clock located there, or to select an action, find the clock to its left. Press the spacebar when the moving hand of your clock is near noon.")

    def gen_side_canvas(self):
	self.h_side = self.h_canvas/2
	self.w_side = self.w_controls
	self.x_side = self.w_canvas
	self.y_side = 0
	self.canvas.create_rectangle((self.x_side, self.y_side, self.x_side+self.w_side, self.y_side+self.h_side), fill=config.bgcolor, outline="black")

    def gen_controls(self):
	self.n_draw = 3
	self.n_toggle = 1
	self.n_meta = 1
	self.canvas.create_rectangle((self.x_controls, self.y_controls, self.x_controls+self.w_controls, self.y_controls+self.n_draw*self.h_control), fill=kconfig.control_bgcolor, outline="black")
	self.canvas.create_rectangle((self.x_controls, self.y_controls+self.n_draw*self.h_control, self.x_controls+self.w_controls, self.y_controls+(self.n_draw+self.n_toggle+self.n_meta)*self.h_control), fill=kconfig.control_bgcolor, outline="black")

	# keys for coloring on wins
        for index in range(self.Nclocks,len(self.rect_locs)):
            self.key_id.append(self.canvas.create_rectangle(self.rect_locs[index],fill=kconfig.key_color,outline=""))

	self.control_text_ids = []
	Ncontrol = len(self.control_text)
	for ncontrol in range(0,Ncontrol):
		self.control_text_ids.append(self.canvas.create_text(self.control_text_locs[ncontrol][0], self.control_text_locs[ncontrol][1], text=self.control_text[ncontrol], anchor=Tkinter.W, font=kconfig.word_font))

    def gen_speed_scale(self):
	self.rotate_index = config.default_rotate_ind
	# well-spaced ticks
	scale_length = self.w_canvas-2*kconfig.clock_rad
	tick_int = int((len(config.period_li)-1)*kconfig.word_pt*3/(1.0*scale_length))+1

	# frame
	self.scale_frame = Tkinter.Frame(self.parent,width=self.w_canvas,height=kconfig.h_scale)
	self.scale_frame.grid(row=1,column=0)
	# scale
	self.scale = Tkinter.Scale(self.scale_frame,orient=Tkinter.HORIZONTAL, label="Clock Rotation Speed", length=scale_length, from_=config.scale_min,to=config.scale_max, tickinterval=tick_int, command=self.change_speed)
	self.speed_index = config.scale_max - self.rotate_index + 1
	self.scale.set(self.speed_index)
	self.scale.pack(side=Tkinter.TOP)
	#******
	self.time_rotate = config.period_li[self.rotate_index]

	# bind up and down arrow events
	self.parent.bind("<Left>",self.dec_speed)
	self.parent.bind("<Right>",self.inc_speed)

	# tool tip
	self.scale_tooltip = ToolTip(self.scale, follow_mouse=1, font=kconfig.word_font, text="This slider scales the speed of clock rotation. Higher values correspond to the clock hand moving faster.")

    def gen_buttons(self):
	# frame for buttons
	self.but_frame = Tkinter.Frame(self.parent,width=self.w_canvas,height=kconfig.h_scale)
	self.but_frame.grid(row=2,column=0)

	# buttons
	self.gen_learn_button()
	
    def gen_learn_button(self):
	self.learn_set = Tkinter.IntVar()
	self.learn_set.set(1)

	# button
	self.learn_button = Tkinter.Checkbutton(self.but_frame, text="Learn", variable = self.learn_set, font=kconfig.word_font, command = self.checked_learn)
	self.learn_button.select()
	self.learn_button.pack(side=Tkinter.LEFT)

	# tool tip
	self.learn_tooltip = ToolTip(self.learn_button, follow_mouse=1, font=kconfig.word_font, text="If this button is checked, the program will adapt to how you click around noon (illustrated in the histogram at right).")

    def checked_learn(self):
	if self.learn_set.get() == 1:
		config.is_learning = True
	else:
		config.is_learning = False


    def gen_histogram(self):
	### histogram
	w_histo = self.w_controls
	h_histo = 2*kconfig.h_scale
	self.histo_canvas = Tkinter.Canvas(self.parent, width=w_histo, height=h_histo, background=config.bgcolor)
	self.histo_canvas.grid(row=1,column=1,rowspan=2)
	bars = self.bc.get_histogram()
	max_height = h_histo *3/4
	norm = max_height * 1.0 / max(bars)
	N_bars = len(bars)
	bar_width = self.w_controls * 1.0 / N_bars

	x = 0
	y = h_histo/8 + max_height
	self.bar_id = []
	for bar in bars:
	    bar_height = bar * norm
	    self.bar_id.append(self.histo_canvas.create_rectangle([x,y,x+bar_width,y-bar_height],fill=kconfig.plot_color,outline=kconfig.plot_outline_color))
	    x += bar_width

	# tooltip to explain what the histogram is
	self.histogram_tooltip = ToolTip(self.histo_canvas, follow_mouse=1, font=kconfig.word_font, text="This is Nomon's estimate of where you click relative to noon on the clocks. The thinner the distribution, the more precisely Nomon thinks you are clicking.")

    def draw_histogram(self):
	w_histo = self.w_controls
	h_histo = 2*kconfig.h_scale

	bars = self.bc.get_histogram()
	max_height = h_histo *3/4
	norm = max_height * 1.0 / max(bars)
	N_bars = len(bars)
	bar_width = self.w_controls * 1.0 / N_bars

	x = 0
	y = h_histo/8 + max_height
	for index in range(0,N_bars):
	    bar_height = bars[index] * norm
	    self.histo_canvas.coords(self.bar_id[index],x,y,x+bar_width,y-bar_height)
	    x += bar_width

    def change_speed(self, index):
	# speed (as shown on scale)
	speed_index = int(index)
	# period (as stored in config.py)
	self.rotate_index = config.scale_max - speed_index + 1
	old_rotate = self.time_rotate
	self.time_rotate = config.period_li[self.rotate_index]
	self.bc.change_period(self.time_rotate)
	self.wait_s = self.bc.get_wait()

    def inc_speed(self, evt):
	index = self.scale.get()
	self.scale.set(min(config.scale_max,index+1))

    def dec_speed(self, evt):
	index = self.scale.get()
	self.scale.set(max(config.scale_min,index-1))

    def on_timer(self):
        if not self.in_pause:
            start_t = time.time()
    
            self.bc.increment(start_t)
		
            self.canvas.focus_set() # so don't have to click on the canvas (e.g. if tooltip takes focus)
    
            self.last_anim_call = self.canvas.after(max(0, int((self.wait_s-(time.time()-start_t))*1000)), self.on_timer)

    def on_press(self, evt):
        if not self.in_pause:
            self.bc.select(time.time())

    def highlight_winner(self, index):
	if (index < self.Nclocks):
		if (self.old_index >= 0):
			self.end_winner(self.old_index)
		self.old_index = index

        self.canvas.itemconfigure(self.key_id[index],fill=kconfig.key_win_color)
        self.canvas.update_idletasks()

    def end_winner(self,index):
	self.canvas.itemconfigure(self.key_id[index],fill=kconfig.key_color)
        self.canvas.update_idletasks()

    def talk_winner(self,talk_string):
        talk_fid = open(self.talk_file,'w')
        talk_fid.write(".. " + talk_string + " ?")
        talk_fid.close()
        os.spawnv(os.P_NOWAIT,'/usr/bin/festival',['festival','--tts',self.talk_file])

    def rectangle_draw(self):
	# find rectangle corners
	ptsA = self.clock_centers[self.loc_inds[-2]]
	ptsB = self.clock_centers[self.loc_inds[-1]]
	if ptsA[0] > ptsB[0]:
		x0 = ptsB[0]
		x1 = ptsA[0]
	else:
		x0 = ptsA[0]
		x1 = ptsB[0]
	if ptsA[1] > ptsB[1]:
		y0 = ptsB[1]
		y1 = ptsA[1]
	else:
		y0 = ptsA[1]
		y1 = ptsB[1]
	# add rectangle
	self.drawn.append(self.canvas.create_rectangle(x0,y0,x1,y1, width=kconfig.line_width, outline=self.line_fill, fill=self.shape_fill))
	Nprev = len(self.drawn)
	if Nprev > 1:
		self.canvas.tag_raise(self.drawn[-1],self.drawn[-2])
	else:
		self.canvas.lower(self.drawn[-1])
	self.side_drawn.append(self.canvas.create_rectangle(self.x_side+x0/2, self.y_side+y0/2, self.x_side+x1/2, self.y_side+y1/2, width=kconfig.side_line_width, outline=kconfig.side_line_fill, fill=self.shape_fill))

    def oval_draw(self):
	# find rectangle corners
	ptsA = self.clock_centers[self.loc_inds[-2]]
	ptsB = self.clock_centers[self.loc_inds[-1]]
	if ptsA[0] > ptsB[0]:
		x0 = ptsB[0]
		x1 = ptsA[0]
	else:
		x0 = ptsA[0]
		x1 = ptsB[0]
	if ptsA[1] > ptsB[1]:
		y0 = ptsB[1]
		y1 = ptsA[1]
	else:
		y0 = ptsA[1]
		y1 = ptsB[1]
	# add oval (bounded by rectangle)
	self.drawn.append(self.canvas.create_oval(x0,y0,x1,y1, width=kconfig.line_width, outline=self.line_fill, fill=self.shape_fill))
	Nprev = len(self.drawn)
	if Nprev > 1:
		self.canvas.tag_raise(self.drawn[-1],self.drawn[-2])
	else:
		self.canvas.lower(self.drawn[-1])
	self.side_drawn.append(self.canvas.create_oval(self.x_side+x0/2, self.y_side+y0/2, self.x_side+x1/2, self.y_side+y1/2, width=kconfig.side_line_width, outline=kconfig.side_line_fill, fill=self.shape_fill))

    def delete_last(self):
	if(len(self.drawn)>0):
		item = self.drawn.pop()
		self.canvas.delete(item)
		item = self.side_drawn.pop()
		self.canvas.delete(item)

	if(self.pen_down == True) and (len(self.loc_inds)>1):
		self.loc_inds.pop()
		self.highlight_winner(self.loc_inds[-1])

    def make_choice(self, index):
        is_undo = False
        is_equalize = False

	if index == self.pen_toggle_clock_ind:
		self.pen_down = True
		self.in_rectangle = False
		self.in_oval = False
		self.loc_inds = []
	elif index == self.rectangle_clock_ind:
		self.pen_down = False
		self.in_rectangle = True
		self.in_oval = False
		self.loc_inds = []
	elif index == self.oval_clock_ind:
		self.pen_down = False
		self.in_rectangle = False
		self.in_oval = True
		self.loc_inds = []
	elif index == self.fill_toggle_clock_ind:
		self.fill_toggle()
	elif index == self.delete_clock_ind:
		self.delete_last()
	elif len(self.loc_inds) == 0:
		self.loc_inds.append(index)
	else:
		self.loc_inds.append(index)
		if self.pen_down:
			self.line_draw()
		elif len(self.loc_inds) == 2:
			if self.in_rectangle:
				self.rectangle_draw()
			if self.in_oval:
				self.oval_draw()
		# otherwise, too many points for the desired shape
		# just don't draw anything

        ## now pause (if desired)
        if kconfig.pause_set:
            self.in_pause = True
            self.canvas.config(bg = kconfig.fill_win_color)
            self.canvas.after(kconfig.pause_length,self.end_pause)
    
        # highlight winner
        self.highlight_winner(index)
	# unhighlight soon if not pixel
	if index >= self.Nclocks:
		self.canvas.after(kconfig.winner_time, self.end_winner, index)
    
        # initialize talk string
        talk_string = self.labels[index]
    
        ## talk the string
        if kconfig.talk_winner_on:
            self.talk_winner(talk_string)

        return (self.words_on, self.words_off, self.word_score_prior, False, False)

    def fill_toggle(self):
	if self.fill_on:
		self.fill_on = False
		self.shape_fill = ""
		new_text = "Turn On Fill"
	else:
		self.fill_on = True
		self.shape_fill = kconfig.shape_fill
		new_text = "Turn Off Fill"

	self.canvas.itemconfigure(self.control_text_ids[self.fill_toggle_control_ind], text=new_text)

    def line_draw(self):

	## draw
	ptA = self.clock_centers[self.loc_inds[-2]]
	ptB = self.clock_centers[self.loc_inds[-1]]
	self.drawn.append(self.canvas.create_line(ptA[0], ptA[1], ptB[0], ptB[1],width=kconfig.line_width,fill=kconfig.line_fill))
	Nprev = len(self.drawn)
	if Nprev > 1:
		self.canvas.tag_raise(self.drawn[-1],self.drawn[-2])
	else:
		self.canvas.lower(self.drawn[-1])
	self.side_drawn.append(self.canvas.create_line(self.x_side+ptA[0]/2, self.y_side+ptA[1]/2, self.x_side+ptB[0]/2, self.y_side+ptB[1]/2, width=kconfig.side_line_width, fill=kconfig.side_line_fill))

    def present_choice(self):
	self.draw_histogram()
        self.canvas.update_idletasks()

    def end_pause(self):
        self.canvas.config(bg=config.bgcolor)
        self.in_pause = False
        self.on_timer()
        
    def quit(self, event=None):
        try: self
        except AttributeError:
            pass
        else:
            ## close clocks
            try: self.bc
            except AttributeError:
                bc_data = []
            else:
                bc_data = self.bc.quit()

        import sys
        sys.exit()

def main():

    root = Tkinter.Tk()
    root.title("Nomon Draw")
    
    app = Keyboard(root)
    root.protocol('WM_DELETE_WINDOW', app.quit)
    root.bind('<Control-q>', app.quit)
    root.bind('<Control-Q>', app.quit)

    root.mainloop()


if __name__ == "__main__":
    main()
